/*    
    Copyright (C) 2012 http://software-talk.org/ (developer@software-talk.org)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package mill.view;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import javax.swing.JPanel;

/**
 * this class provides the main panel for the game "Mill".
 * It contains the background of the gamefield.
 * @version 1.2.0
 */
public class MainPanel extends JPanel {

    private Color color;
    private Color bgColor;
    private int thickness;
    private int width;
    private int lineDistance;
    private int borderDistance;
    private boolean playWithDiagonalLines;

    /**
     * instance a new <code>MainPanel</code> with the standard backgroundcolor
     * yellow and the standard color for painting the lines black.
     * The background will be set to the backgroundcolor.
     */
    public MainPanel() {
        super();
        color = Color.BLACK;
        bgColor = new Color(139, 69, 19);
        playWithDiagonalLines = true;
        super.setBackground(bgColor);
    }
    /**
     * sets if the diagonal lines should be draw.
     * @param playWithDiagonalLines 
     */
    public void setPlayWithDiagonalLines(boolean playWithDiagonalLines) {
        this.playWithDiagonalLines = playWithDiagonalLines;
    }
 
    
    

    /**
     * to set the color of the lines. Default it will be <Code>Color.Black</code>.
     * @param color the color of the lines
     */
    public void setColor(Color color) {
        this.color = color;
    }

    @Override
    public void setBackground(Color color) {
        super.setBackground(color);
        this.bgColor = color;
    }

    @Override
    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        width = getSize().width;
        calcBorderDistance();
        width = width - 2 * borderDistance;
        calcThickness();
        calcLineDistance();
        drawSquares(g);
        drawLines(g);
        if (playWithDiagonalLines) {
            drawDiagonalLines(g);
        }
    }

    /**
     * calcs the thickness of the lines. In this case it is
     * 2% of the width.
     */
    private void calcThickness() {
        thickness = (int) (width * 0.02);
    }

    /**
     * calcs the distance from on line to the other. In the case it is
     * a six part of the widht rounding down.
     */
    private void calcLineDistance() {
        lineDistance = (int) (width / 6);
    }

    /**
     * calcs the breadth of the border. In this case the border will be
     * the distance from the edge of the panel to the first painted line.
     * It will be 6% of the width. Also the painted area is 12% smaller then
     * the width of the panel.
     */
    private void calcBorderDistance() {
        borderDistance = (int) (width * 0.06);
    }

    /**
     * draws the four single lines of the gamefield. This are the lines
     * from the middle border to the centersquare.
     * @param g the graphics object to draw to
     */
    private void drawLines(Graphics g) {
        int x1;
        int x2;
        int y1;
        int y2;
        Graphics2D g2D = (Graphics2D) g;
        g2D.setColor(color);
        g2D.setStroke(new BasicStroke(thickness));
        // the line from top to middle
        x1 = width / 2 + borderDistance;
        y1 = borderDistance + thickness;
        x2 = x1;
        y2 = borderDistance + 2 * lineDistance;
        g2D.drawLine(x1, y1, x2, y2);
        
       // the line from bottom to middle
        y1 = borderDistance + 4 * lineDistance;
        y2 = width + borderDistance - thickness;
        g2D.drawLine(x1, y1, x2, y2);
       // the line from left to middle
        x1 = borderDistance + thickness;
        x2 = 2 * lineDistance + borderDistance;
        y1 = borderDistance + 3 * lineDistance;
        y2 = y1;
        g.setColor(color);
        g2D.drawLine(x1, y1, x2, y2);
        // the line from right to middle
        x1 = width + borderDistance - thickness;
        x2 = width - 2 * lineDistance + borderDistance;
        y1 = borderDistance + 3 * lineDistance;
        y2 = y1;
        g2D.drawLine(x1, y1, x2, y2);
    }

    /**
     * draws three quadrates for the gamefield. The three quadrates are
     * the outer squareline, the middle squareline, the inner squareline.
     * It will be painted in this order and each painted square has
     * a square inside with the backgroundcolor, which is a little bit smaller
     * addicted from the thickness of the lines; this it to make thickness lines.
     * @param g the graphic object to paint to
     */
    private void drawSquares(Graphics g) {
        // first square
        int x = borderDistance;
        int y = borderDistance;
        int currentWidth = width;
        int currentHeight = width;
        g.setColor(color);
        g.fillRect(x, y, currentWidth, currentHeight);
        // 
        g.setColor(bgColor);
        x += thickness;
        y += thickness;
        currentWidth -= 2 * thickness;
        currentHeight -= 2 * thickness;
        g.fillRect(x, y, currentWidth, currentHeight);
        // second square
        g.setColor(color);
        x = borderDistance + lineDistance;
        y = borderDistance + lineDistance;
        currentWidth = width - 2 * lineDistance;
        currentHeight = width - 2 * lineDistance;
        g.fillRect(x, y, currentWidth, currentHeight);

        g.setColor(bgColor);
        x += thickness;
        y += thickness;
        currentWidth -= 2 * thickness;
        currentHeight -= 2 * thickness;
        g.fillRect(x, y, currentWidth, currentHeight);
        // third square
        g.setColor(color);
        x = borderDistance + 2 * lineDistance;
        y = borderDistance + 2 * lineDistance;
        currentWidth = width - 4 * lineDistance;
        currentHeight = width - 4 * lineDistance;
        g.fillRect(x, y, currentWidth, currentHeight);

        g.setColor(bgColor);
        x += thickness;
        y += thickness;
        currentWidth -= 2 * thickness;
        currentHeight -= 2 * thickness;
        g.fillRect(x, y, currentWidth, currentHeight);

    }

    private void drawDiagonalLines(Graphics g) {
        Graphics2D g2D =(Graphics2D) g;
        g2D.setColor(color); 
        // set up the stroke:
       g2D.setStroke(new BasicStroke(thickness));
        // upper left to middle
        int x = borderDistance + thickness;
        int y = borderDistance + thickness;
        int x2 = borderDistance + 2 * lineDistance;
        int y2 = borderDistance + 2 * lineDistance;
        g2D.drawLine(x, y, x2, y2);

        // upper right to middle
        x = width + borderDistance - thickness;
        y = borderDistance + thickness;
        x2 = width - 2 * lineDistance + borderDistance; //(borderDistance + 2 * lineDistance) + (width - 4 * lineDistance);
        y2 = width - 4 * lineDistance + borderDistance;
        g2D.drawLine(x, y, x2, y2);
        
        // lower left to middle
        x = borderDistance + thickness;
        y = width + borderDistance - thickness;
        x2 = borderDistance + 2 * lineDistance;
        y2 = (borderDistance + 2 * lineDistance) + (width - 4 * lineDistance);    
        g2D.drawLine(x, y, x2, y2);

        // lower right to middle
        x = borderDistance + width - thickness;
        y = borderDistance + width - thickness;
        x2 = (borderDistance + 2 * lineDistance) + (width - 4 * lineDistance);
        y2 = (borderDistance + 2 * lineDistance) + (width - 4 * lineDistance);
        g2D.drawLine(x, y, x2, y2);

    }
}
